# Update {{ connector_technical_name }}

This PR was autogenerated by running `airbyte-ci connectors --name={{ connector_technical_name }} up_to_date --pull`

We've set the `auto-merge` label on it, so it will be automatically merged if the CI pipelines pass.
If you don't want to merge it automatically, please remove the `auto-merge` label.
Please reach out to the Airbyte Connector Tooling team if you have any questions or concerns.

{% if step_results %}
## Operations
{% for step_result in step_results %}
- {{step_result.step.title}}: {{ step_result.status.value }}
{% endfor %}
{% endif %}

{% if dependency_updates %}
## Dependency updates

We use [`syft`](https://github.com/anchore/syft) to generate a SBOM for the latest connector version and the one from the PR.
It allows us to spot the dependencies that have been updated at all levels and for all types of dependencies (system, python, java etc.).
Here are the dependencies that have been updated compared to `{{ connector_latest_docker_image }}`.
Keep in mind that `:latest` does not always match the connector code on the main branch. 
It is the latest released connector image when the head commit of this branch was created.

| Type | Name | State | Previous Version | New Version |
|------|------|-------|-------------|------------------|
{%- for dependency_update in dependency_updates %}
| {{ dependency_update.package_type }} | {{ dependency_update.package_name }} | {{ dependency_update.update_type.value }} | {{ dependency_update.previous_version }} | **{{ dependency_update.new_version }}** |
{%- endfor %}
{% endif %}


